# Copyright (c) 2012 David Miller (david@deadpansincerity.com)
#
# This file is part of zoop (http://github.com/davidmiller/zoop)
#
# zoop is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
zoop.tree

Representing ZooKeeper Node trees.

>>> t = Tree()
>>> t.add_nodes('/foo/bar/goo/car')
"""
class Tree(object):
    def __init__(self):
        self.nodes = {}

    def __repr__(self):
        return "<zoop Tree>"

    @staticmethod
    def from_zk(zk, root='/', level=None, pattern=None, exclude=None):
        """
        Inspect the ZooKeeper instance connected in `zk, and create a
        tree representing the schema starting at `root`, and descending
        `level` subdirectories.

        If the `pattern` argument is passed, only include those Nodes whose
        name matches the regexp `pattern`.

        If the `exclude` argument is passed, exclude any Nodes whose name
        matches the regexp `exclude`

        Arguments:
        - `zk`: ZooKeeper
        - `root`: str - the node to start at.
        - `level`: int - the number of child Nodes to descend.
        - `pattern`: str - regexp
        - `exclude`: str - regexp

        Return: Tree
        Exceptions: None
        """
        return Tree()

    def add_nodes(self, path, nodes):
        # This is very possibly not the API you are looking for
        pathlist = [p for p in path.split('/') if p]
        print pathlist
        current = self.nodes
        for path in pathlist:
            if not path in current or current[path] == None:
                current[path] = {}
            current = current[path]
        for n in nodes:
            current[n] = None

    def tree(self):
        """
        Return a string representation of the Nodes
        in this ZooKeeper Instance

        Return: str
        Exceptions: None
        """
        lines = ['/']

        def rectree(nodes, prefix):
            nodenames = sorted(nodes.keys())
            for i, child in enumerate(nodenames):
                lines.append('{0}|-- {1}'.format(prefix, child))
                if nodes[child]:
                    nodefix = ' ' * 4
                    if i + 1 < len(nodenames):
                        nodefix = '|{0}'.format(nodefix[1:])
                    subfix = '{0}{1}'.format(prefix, nodefix)
                    rectree(nodes[child], subfix)

        rectree(self.nodes, ' ')
        return "\n".join(lines)

    def pprint(self):
        """
        Print a string representation of the Nodes
        in this ZooKeeper Instance

        Return: None
        Exceptions: None
        """
        print self.tree()
        return




"""
Make us a package  please!
"""
from zoop._version import __version__
from zoop import exceptions
from zoop.client import ZooKeeper
from zoop.enums import Event
from zoop.lock import Lock
from zoop.logutils import divert_zoolog
from zoop.queue import Queue
from zoop.tree import Tree

__all__ = [
    'exceptions',
    'ZooKeeper',
    '__version__',
    'divert_zoolog',
    'Event',
    'Lock',
    'Queue',
    'Tree'
    ]



# Copyright (c) 2012 David Miller (david@deadpansincerity.com)
#
# This file is part of zoop (http://github.com/davidmiller/zoop)
#
# zoop is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
zoop.queue

Distributed Queue implementation running on Zookeeper.

As much as is possible this implementation seeks to mirror the API
established by the Standard Library's queue module.

"""
import os

import zookeeper

from zoop import enums, exceptions, lock

class Queue(object):
    """
    A FIFO queue for ZooKeeper.

    Specifying the prefix allows you interoperability with
    Queue implementations from other libraries.

    Arguments:
    - `client`: ZooKeeper
    - `path`: string Path we want to treat as a queue
    - `prefix`: prefix string for the item nodes.

    >>> zk = ZooKeeper('localhost:2181')
    >>> myq = Queue(zk, '/myq')
    >>> myq.put("Frist")
    >>> myq.put("Next")
    >>> myq.get()
    "Frist"
    >>> myq.get()
    "Next"
    >>> myq.get()
    Traceback (most recent call last):
        ...
    Empty: No items in queue at /myq
    """
    def __init__(self, client, path, prefix='q-'):
        self.zk = client
        self.path = path
        self.prefix = prefix
        if not self.zk.exists(path):
            self.zk.create(path)
        name = os.path.basename(path) + '-lock'
        self.lock = lock.Lock(client, name, os.path.dirname(path))

    def __repr__(self):
        return "<ZooKeeper FIFO Queue at {0}{1}>".format(self.zk.server, self.path)

    def empty(self):
        """
        Return ``True`` if the queue is empty, False otherwise.
        Note, empty() == False does not guarantee that a subsequent
        get() will not raise an Empty error.

        Return: bool
        Exceptions: None
        """
        return len(self.zk.get_children(self.path)) == 0

    def flush(self):
        """
        Flush the Queue's current state, deleting all nodes.

        Return: None
        Exceptions: None
        """
        kids = self.zk.get_children(self.path)
        for k in kids:
            self.zk.delete(os.path.join(self.path, k))
        return

    def get(self):
        """
        Return the next item from the Queue

        Return: string data item
        Exceptions: Empty
        """
        frist = self.sorted()[0] # This can raise Empty()
        ipath = os.path.join(self.path, frist)
        item = self.zk.get(ipath)
        self.zk.delete(ipath)
        return item

    def put(self, item):
        """
        Put `item` at the end of the queue.

        Arguments:
        - `item`: string - data to add

        Return: None
        Exceptions: None
        """
        return self.zk.create(os.path.join(self.path, self.prefix),
                              value=item,
                              flags=zookeeper.SEQUENCE)

    def qsize(self):
        """
        Return the size of the queue.
        Note, qsize() > 0 doesn't guarantee that a subsequent
        get() will not raise an Empty error.

        The watch() API is the recommended way to consume items
        from a queue.

        Return: int
        Exceptions: None
        """
        return len(self.zk.get_children(self.path))

    def sorted(self):
        """
        Return the items in the Queue, sorted by time added

        Return: list containing strings
        Exceptions: Empty
        """
        contents = self.zk.get_children(self.path)
        if not contents:
            raise exceptions.Empty("Queue at {0} has no items".format(self.path))
        return sorted(contents)

    def watch(self, callback):
        """
        Watch the Queue for items being added, and when they are,
        call `callback` with a list of items currently in the Queue.

        `callback` should be a callable that takes a single argument.
        It should expect a list of strings.

        Arguments:
        - `callback`: callable

        Return: None
        Exceptions: None

        >>> zk = ZooKeeper('localhost:2181')
        >>> zk.connect()
        >>> myq = Queue(zk, '/myq')
        >>> def watchit(data):
        ...     print "Watchit got", data, "!"
        >>> assert(callable(watchit))
        >>> myq.watch(watchit)
        >>> myq.put("Frist")
        Watchit got ['q-00000001'] !
        """
        def watcher(event, path):
            """
            Watch for items added to the Queue, then remove the latest and
            run the callback on it.
            """
            try:
                kids = self.sorted()
            except exceptions.Empty:
                return # Deleted event
            return callback(kids)

        self.zk.watch(self.path, watcher, enums.Event.Child)
        return

    def watchitem(self, callback):
        """
        Watch for items being added to the queue, and call
        callback with the contents of added nodes when they
        are added.

        Callback should be callable and should take one
        argument, the data for the item that has just been
        added to the queue.

        Arguments:
        - `callback`: callable

        Return: None
        Exceptions: None

        >>> zk = ZooKeeper('localhost:2181')
        >>> zk.connect()
        >>> myq = Queue(zk, '/myq')
        >>> def watchit(data):
        ...     print "Watchit got", data, "!"
        >>> assert(callable(watchit))
        >>> myq.watchitem(watchit)
        >>> myq.put("Frist")
        Watchit got "Frist" !
        """
        def watcher(event, path):
            """
            Watch for items added to the Queue, then remove the latest and
            run the callback on it.
            """
            try:
                data, stats = self.get()
            except exceptions.Empty:
                return # Deleted event
            return callback(data)

        self.zk.watch(self.path, watcher, enums.Event.Child)



"""
!!! PriorityQueue

To implement a priority queue, you need only make two simple changes
to the generic queue recipe . First, to add to a queue, the pathname
ends with "queue-YY" where YY is the priority of the element with lower
numbers representing higher priority (just like UNIX). Second, when
removing from the queue, a client uses an up-to-date children list
meaning that the client will invalidate previously obtained children
lists if a watch notification triggers for the queue node.
"""



"""
zoop.logutils

ZooKeepr has the irritating quality of printin it's logs to sterr

Sometimes (mostly) this is not really a desirable thing
"""
import zookeeper

def divert_zoolog(logfile='/dev/null'):
    """
    Divert the ZooKeeper logging away from stderr.

    This is particularly useful in interactive sessions where
    the default logging behaviour is distinctly disruptive, but
    also useful when logging that output to a file is desired.


    Arguments:
    - `logfile`: string - path to the file we'll log to

    Return: None
    Exceptions: None
    """
    fh = open(logfile, 'w')
    zookeeper.set_log_stream(fh)
    return

def set_loglevel(level):
    """
    Set the ZooKeeper log level.

    Valid arguments are DEBUG|ERROR|INFO|WARN


    Arguments:
    - `level`: string

    Return: None
    Exceptions: None
    """
    zlevel = getattr(zookeeper, 'LOG_LEVEL_{0}'.format(level))
    zookeeper.set_debug_level(zlevel)
    return



"""
Enums.

Nicer Python mappings to the ZooKeeper constants
"""
import zookeeper

class Event(object):
    "Enum For ZooKeeper Event types"
    NotWatching = zookeeper.NOTWATCHING_EVENT
    Session = zookeeper.SESSION_EVENT
    Created = zookeeper.CREATED_EVENT
    Deleted = zookeeper.DELETED_EVENT
    Changed = zookeeper.CHANGED_EVENT
    Child   = zookeeper.CHILD_EVENT

    @classmethod
    def human(cls, item):
        """
        Given an integer representing an item in this Enum,
        fetch us the name of that property.


        Arguments:
        - `item`: int

        Return: str
        Exceptions: None
        """
        vals = cls.__dict__.values()
        thisone = [v for k, v in vals if v == item]
        return thisone[0]



# Copyright (c) 2012 David Miller (david@deadpansincerity.com)
#
# This file is part of zoop (http://github.com/davidmiller/zoop)
#
# zoop is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
zoop.lock

A distributed Lock on top of ZooKeeper
"""
from os.path import join
import threading
import time

import zookeeper

from zoop import exceptions

class BaseLock(object):
    """
    A base for all subsequent locks to inherit from.
    """
    prefix = 'baselock-'

    def __init__(self, handle, name, root='/zooplocks'):
        """
        Store instance vars and ensure that the Node exists

        Arguments:
        - `handle`: ZooKeeper
        - `name`: str
        - `root`: str

        Return: None
        Exceptions: None
        """
        self.zk = handle
        self.name = name
        self.root = root
        self.path = join(root, name)
        self.tlocal = threading.local()
        self.tlocal.revoked = []
        self.tlocal.locking = None
        self.tlocal.acquired = False
        if not self.zk.exists(root):
            self.zk.create(root)
        if not self.zk.exists(self.path):
            self.zk.create(self.path)
        return

    def __repr__(self):
        return "<BaseLock for {0}>".format(self.path)

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    @property
    def revoked(self):
        """
        Predicate to indicate whether we have been asked to
        release this lock.

        Returns: bool
        Exceptions: None
        """
        return bool(self.tlocal.revoked)

    def acquire(self, timeout=None):
        """
        Attempt to acquire the lock.

        If a timeout parameter is passed, only wait this long
        for acquisition.

        Arguments:
        - `timeout`: int

        Return: bool - whether we acquired the Lock or not
        Exceptions: None
        """
        # This implementation is based upon the Mozilla Services
        # ztools lock at https://github.com/mozilla-services/zktools
        # with some additional encapsulation and error handling added.
        self.tlocal.revoked = []

        nodepath, keyname = self._create_waitnode()
        cv = threading.Event()

        def lockwatch(handle, etype, state, path):
            cv.set()

        tstart = time.time()
        frist = True
        while not self.tlocal.acquired:
            cv.clear()

            if not frist and timeout is not None and time.time() - tstart > timeout:
                try:
                    self.zk.delete(nodepath)
                except exceptions.NoNodeError:
                    pass
                return False
            frist = False

            kids = self.zk.get_children(self.path)

            if len(kids) == 0 or not keyname in kids:
                # Only really for connection issues
                nodepath, keyname = self._create_waitnode()
                continue

            self.acquired, blocking = self.has_lock(keyname, kids)
            if self.acquired:
                break

            last_blocker = join(self.path, blocking[-1])
            if not self.zk.exists(last_blocker, lockwatch):
                continue # Already free

            if timeout is not None:
                cv.wait(timeout - (time.time() - tstart))

        self.tlocal.lock_node = nodepath
        return True

    def _create_waitnode(self):
        """
        Create the wait node for our thread.
        When this is the lowest number, we have the lock.

        Return: tuple of strings -
               * Full path of the child
               * The nodename of the child
        Exceptions: None
        """

        def revoked(handle, etype, conn, path):
            removes = [zookeeper.DELETED_EVENT, zookeeper.EXPIRED_SESSION_STATE]
            if etype in removes:
                try:
                    self.tlocal.revoked.append(True)
                except AttributeError:
                    pass # We've already unlocked
                return

            if etype == zookeeper.CHANGED_EVENT:
                data = self.zk.get(path, revoked)
                if data == 'unlock':
                    try:
                        self.tlocal.revoked.append(True)
                    except AttributeError:
                        pass # We've already unlocked
            return

        nodepath = self.zk.create(join(self.path, self.prefix), value="0",
                                  flags=zookeeper.SEQUENCE)
        data = self.zk.get(nodepath, watch=revoked)[0]
        if data == 'unlock':
            self.tlocal.revoked.append(True)

        keyname = nodepath.split('/')[-1]
        return nodepath, keyname

    def has_lock(self, keypath, locknodes):
        """
        Determine whether the current Thread has the Lock.
        Look to see whether we're frist in the current locknodes.

        If we are, then great! - we have achieved Lock-varna.

        Otherwise, all locknodes before ours are blocking us :(

        Arguments:
        - `keypath`: string - this thread's node
        - `locknodes`: list of strings all wait nodes

        Return: tuple of (bool, list or None)
                we return a boolean which represents whether or
                not we have the lock, and a list of blocking Nodes
                or None.
        Exceptions: None
        """
        locknodes.sort(key=lambda val: val[val.rfind('-') + 1:])

        if keypath == locknodes[0]:
            return True, None
        return False, locknodes[:locknodes.index(keypath)]

    def release(self):
        """
        Release a Lock!

        Return: None
        Exceptions: None
        """
        self.tlocal.revoked = []
        self.acquired = False
        try:
            self.zk.delete(self.tlocal.lock_node)
            del self.tlocal.lock_node
        except (zookeeper.NoNodeException, AttributeError):
            pass # We never had the Lock!
        return True

class Lock(BaseLock):
    """
    Synchronise your lockage!

    >>> zk = ZooKeeper('localhost:2181')
    >>> zk.connect()
    >>> lk = Lock(zk, 'mylock')
    >>> with lk:
    ...     print zk.get_children('/zooplocks/mylock')
    ...
    ['lock-0000001']
    >>> zk.get_children('/zooplocks/mylock')
    []
    """
    prefix = 'lock-'



# Copyright (c) 2012 David Miller (david@deadpansincerity.com)
#
# This file is part of zoop (http://github.com/davidmiller/zoop)
#
# zoop is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
zoop.watcher

Callbacks for state change go here!
"""
import collections
import threading

PLock = threading.RLock()

import zookeeper

from zoop import enums, exceptions

class Watcher(object):
    """
    Stores a register of callbacks for particular watchers.

    >>> zk = zookeeper.init('localhost:2181)
    >>> watcher = Watcher(zk)
    >>> def cb(event, path):
    ...     print 'Got Called!", path, event
    ...
    >>> watcher.spyon('/zookeeper', cb, Event.Child)
    >>> client = Client('localhost:2181')
    >>> client.create('/zookeeper/yay', 'Hello Beautiful World')
    ... Got Called! /zookeeper/yay 4
    """

    _watch_funcs = {
        enums.Event.Deleted: zookeeper.get,
        enums.Event.Changed: zookeeper.get,
        enums.Event.Child: zookeeper.get_children
        }

    def __init__(self, zkh):
        """
        Store vars

        Arguments:
        - `zkh`: zookeeper instance handle

        Return: None
        Exceptions: None
        """
        self._zk = zkh
        self.callbacks = collections.defaultdict(lambda: collections.defaultdict(list))

    def set_zhandle(self, handle):
        """
        Attribute setter for the watcher's active handle
        to a ZooKeeper instance.

        Although this is passed in at instantiation, we also
        sometimes want an interface to change the handle,
        without having calling code understand the internal
        implementation details of the Watcher class.

        Arguments:
        - `handle`: ZooKeeper instance handle

        Return: None
        Exceptions: None
        """
        self._zk = handle
        return

    def set_global(self):
        """
        Set our dispatch function as the ZooKeeper global watcher.

        Return: None
        Exceptions: None
        """
        zookeeper.set_watcher(self._zk, self.dispatch)
        return

    def dispatch(self, zk, etype, conn, path):
        """
        Callback for libzookeeper that fires when ZooKeeper events occur.

        Dispatch to our own callbacks.
        Then re-watch the node/event type

        Arguments:
        - `zk`: handle to the ZooKeeper connection
        - `etype`: Enum- Event type
        - `conn`: Enum Connection Status
        - `path`: string- Path that the event occured

        Return: None
        Exceptions: None

        """
        if etype == enums.Event.Session:
            return

        if etype in self._watch_funcs:
            self._watch_funcs[etype](self._zk, path, self.dispatch)

        if self.callbacks[path][etype] > 0:
            for cb in self.callbacks[path][etype]:
                cb(path, etype)

        def watchit(h, t, s, p):
            self.dispatch(h, t, s, p)

        return

    def spyon(self, path, callback, *events):
        """
        Begin watching `path` for events of type `event`.
        When one happens, execute `callback`, with two
        arguments, the path of the ZooKeeper Event and the Event type.

        Arguments:
        - `path`: string - Path to watch
        - `callback`: callable
        - `*events`: int - one or more zoop.Event attribute. Must pass
                           at least one

        Return: None
        Exceptions:
        - NoEventError: No events passed.
        """
        if not events: # Valid syntax, but invalid semantics
            errmsg = "You must pass at least one Event to spy on"
            raise exceptions.NoEventError(errmsg)
        for e in events:
            self.callbacks[path][e].append(callback)

            def cb(h, t, s, p):
                self.dispatch(h, t, s, p)
                return

            print self._watch_funcs[e](self._zk, path, cb)
        return



# Copyright (c) 2012 David Miller (david@deadpansincerity.com)
#
# This file is part of zoop (http://github.com/davidmiller/zoop)
#
# zoop is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
zoop.client
"""
from os.path import join
import threading

import zookeeper

from zoop import exceptions, watch

OPEN_ACL_UNSAFE = dict(perms=zookeeper.PERM_ALL, scheme = 'world', id='anyone')

class BaseZK(object):
    """
    Common ZooKeeper Client protocol for subclassing
    """
    flavour = 'Base Client'

    def __init__(self, connection):
        """
        Create the zookeeper.client instance

        Arguments:
        - `connection`: string host:port
        """
        self.connwait = 15.0
        self.connected = False
        self.cv = threading.Condition()
        self.server = connection
        self._zk = None
        self.watcher = watch.Watcher(self._zk)
        return

    def __repr__(self):
        return "<ZooKeeper {0} for {1}>".format(
            self.flavour, self.server)

    def connect(self):
        """
        Create a connection to the ZooKeeper instance

        Return: None
        Exceptions: None
        """
        def connwatch(handle, etype, state, path):
            with self.cv:
                self.connected = True
                self.cv.notify()
            return

        with self.cv:
            self._zk = zookeeper.init(self.server, connwatch)
            self.cv.wait(self.connwait)
            if not self.connected:
                raise Exception("!")

        self.watcher.set_zhandle(self._zk)
        self.watcher.set_global()

    def close(self):
        """
        Close the connection to our ZooKeeper instance

        Return: None
        Exceptions: None
        """
        if self._zk:
            zookeeper.close(self._zk)

    """
    Here we start stub methods that define the client API that remains
    constant throughout subclasses.
    """

    def create(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def delete(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def exists(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def get(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def get_children(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def set(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def watch(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    """
    The following are either aliases, or generic abstractions that
    rely on the implementation of the APIs above.
    """

    def ls(self, path):
        """
        Return a list of strings representing the child nodes of `path`

        (Note, this is an alias of get_children)
        Arguments:
        - `path`: string

        Return: list of strings
        Exceptions: NoNodeError
        """
        return self.get_children(path)

    def rm(self, path):
        """
        Delete the node at `path`

        (Note, this is an alias of delete)
        Arguments:
        - `path`: string

        Return: None
        Exceptions: None
        """
        return self.delete(path)

    def mkdirp(self, path):
        """
        Recursively make all nodes in the given path

        Arguments:
        - `path`: string

        Return: None
        Exceptions: None
        """
        paths = path.split('/')
        for i, path in enumerate(paths):
            p = join(join(paths[:i]), path)
            if not self.exists(p):
                self.create(p)

    def rm_rf(self, path):
        """
        Recursively delete all nodes below the given path

        Arguments:
        - `path`: string

        Return: None
        Exceptions: None
        """
        kids = self.ls(path)
        if kids:
            for k in kids:
                self.rm_rf(join(path, k))
        return self.rm(path)

    """
    Factory methods to return objects that require an instance of the
    client, ambivalent to what flavour of client they are actually dealing with.
    """

    def Queue(self, path, prefix='q-'):
        """
        Returns an instantiated Queue with the root `path`
        connected to this ZooKeeper instance.

        Arguments:
        - `path`: string - the root of your Queue. should be an
                           absolute path.
        - `prefix`: prefix string for the item nodes.

        Return: Queue
        Exceptions: NotConnectedError - the ZooKeeper instance was not connected
        """
        if not self.connected:
            err = "You aren't connected to a ZooKeeper instance - no way to create a Queue"
            raise exceptions.NotConnectedError(err)
        # Avoid circular imports from the top-level package namespace
        from zoop import queue
        return queue.Queue(self, path, prefix=prefix)


class ZooKeeper(BaseZK):
    """
    The ZooKeeper client

    Arguments:
    - `connection`: string of host:port

    Return: None
    Exceptions: None
    """
    flavour = 'Client'

    def create(self, path, value='', acl=[OPEN_ACL_UNSAFE], flags=0):
        """
        Create a new Node at `path` containing `value` on our ZooKeeper instance.

        Arguments:
        - `path`: string - new path
        - `value`: string - value of the Node
        - `acl`: list - list of Access Control flags
        - `flags`: int - the ZooKeeper flags (SEQUENCE|EPHEMERAL)

        Return: None
        Exceptions:
        - NodeExistsError: This Node already exists
        - NoNodeError: A parent Node in `path` does not exist

        """
        try:
            return zookeeper.create(self._zk, path, value, acl, flags)
        except zookeeper.NodeExistsException:
            errstr = "Can't create {0} as it already exists".format(path)
            raise exceptions.NodeExistsError(errstr)
        except zookeeper.NoNodeException:
            errstr = "A parent node of {0} does not exist".format(path)
            raise exceptions.NoNodeError(errstr)

    def delete(self, path):
        """
        Delete the ZooKeeper Node at `path`

        Arguments:
        - `path`: string

        Return: None
        Exceptions: NoNodeError
        """
        try:
            zookeeper.delete(self._zk, path)
        except zookeeper.NoNodeException:
            errmsg = "The Node {0} does not exist".format(path)
            raise exceptions.NoNodeError(errmsg)

    def exists(self, path, watch=None):
        """
        Determine whether the ZooKeeper Node at `path` exists

        Arguments:
        - `path`: string

        Return: dict of stats or None
        Exceptions: None
        """
        return zookeeper.exists(self._zk, path, watch)

    def get(self, path, watch=None):
        """
        Get the value of the ZooKeeper Node at `path`

        Arguments:
        - `path`: string
        - `watch`: callable - optional watcher function

        Return: Tuple of (Value, Statsdict)
        Exceptions: NoNodeError
        """
        return zookeeper.get(self._zk, path, watch)

    def get_children(self, path):
        """
        Return a list of strings representing the child nodes of `path`

        Arguments:
        - `path`: string

        Return: list of strings
        Exceptions: NoNodeError
        """
        # !!! Wrap the ZooKeeper exceptions
        return zookeeper.get_children(self._zk, path, None)

    def set(self, path, value):
        """
        Set the value of the ZooKeeper Node at `path`

        Arguments:
        - `path`: string
        - `value`: string

        Return: None
        Exceptions: NoNodeError
        """
        zookeeper.set(self._zk, path, value)
        return

    def watch(self, path, callback, event):
        """
        Begin watching `path` for events of type `event`.
        When one happens, execute `callback`, with two
        arguments, the path of the ZooKeeper Even and the event type

        Arguments:
        - `path`: string - Path to watch
        - `callback`: callable
        - `event`: int - a zoop.Event attribute

        Return: None
        Exceptions: None
        """
        self.watcher.spyon(path, callback, event)
        return


class AsyncZooKeeper(BaseZK):
    """
    A ZooKeeper client that uses the Asynchronous
    libzookeeper API.

    Methods can be expected to take an additional
    callback parameter.
    """
    flavour = 'Async Client'

    def create(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def delete(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def exists(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def get(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def get_children(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """

        raise NotImplementedError("!")

    def set(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")

    def watch(self, *a, **kw):
        """
        This is a method stub for subclasses to override.

        Return: None
        Exceptions: NotImplementedError
        """
        raise NotImplementedError("!")



"""
Exceptions for the zoop library
"""
class Error(Exception):
    "Base Error from which all zoop exceptions must inherit"

class NoEventError(Error):
    "We wanted an event, but didn't get one."

class LostConnectionError(Error):
    "Lost connection to the ZooKeeper instance"

class NotConnectedError(Error):
    "Not Connected to a ZooKeeper instance."

class NodeExistsError(Error):
    "The Node Exists!"

class NoNodeError(Error):
    "This Node does not exist"

class Empty(Error):
    "The item in question is empty."



"""
unittests for the zoop.lock module
"""
import sys
import unittest
if sys.version_info < (2, 7):
    import unittest2 as unittest

from mock import patch, Mock
import zookeeper

from zoop import lock

class BaseLockTestCase(unittest.TestCase):
    def setUp(self):
        self.zk = Mock(name='Mock ZooKeeper')
        self.lk = lock.BaseLock(self.zk, 'barlock')

    def test_init(self):
        """ Initializer """
        self.zk.exists.return_value = False

        lk = lock.BaseLock(self.zk, 'foolock', root='/lockz')
        self.assertEqual('/lockz/foolock', lk.path)
        self.zk.exists.assert_any_call('/lockz/foolock')
        self.zk.exists.assert_any_call('/lockz')
        self.zk.create.assert_any_call('/lockz/foolock')
        self.zk.create.assert_any_call('/lockz')

    def test_contextmanager(self):
        "Can we use it as a contextmanager?"
        with patch.object(self.lk, 'acquire') as Pac:
            with patch.object(self.lk, 'release') as Prel:
                with self.lk:
                    Pac.assert_called_once_with()
                Prel.assert_called_once_with()

    def test_revoked(self):
        """ Predicatise a list """
        self.lk.tlocal.revoked.append(True)
        self.assertEqual(True, self.lk.revoked)

    def test_not_revoked(self):
        """ We haven't been asked to revoke yet. """
        self.assertEqual(False, self.lk.revoked)

    def test_acquire(self):
        "Acquire the lock"
        # !!! This is brittle.
        self.zk.get.return_value = 'got'
        self.zk.create.return_value = '/lockz/foolock/lock-00000001'
        self.zk.get_children.return_value = ['lock-00000001']
        self.assertEqual(True, self.lk.acquire())

    def test_create_waitnode(self):
        "Create a wait node."
        self.zk.create.return_value = '/zooplocks/barlock/baselock-00000001'
        self.zk.get.return_value = 'got'

        nodepath, keynode = self.lk._create_waitnode()
        self.zk.create.assert_called_once_with('/zooplocks/barlock/baselock-',
                                               value = '0',
                                               flags = zookeeper.SEQUENCE)
        self.assertEqual('/zooplocks/barlock/baselock-00000001', nodepath)
        self.assertEqual('baselock-00000001', keynode)

    def test_has_lock(self):
        """ Do we have the lock """
        cases = [
            ((True, None), ('baselock-0001', ['baselock-0001', 'baselock-0002'])),
            ((False, ['baselock-0001']), ('baselock-0002', ['baselock-0001', 'baselock-0002']))
            ]
        for expected, arg in cases:
            actual = self.lk.has_lock(*arg)
            self.assertEqual(expected, actual)

    def test_release(self):
        "Can we release the lock?"
        self.assertEqual(True, self.lk.release())

class LockTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_init(self):
        """ Initializer """
        pass

    def tearDown(self):
        pass



if __name__ == '__main__':
    unittest.main()



"""
Unittests for the zoop.tree module
"""
import sys
import unittest
if sys.version_info < (2, 7):
    import unittest2 as unittest

from mock import patch

from zoop import tree

SIMPLE_TREE = """/
 |-- foo
 |   |-- bar
 |   |-- car
 |-- goo"""

NESTED_TREE = """/
 |-- foo
 |   |-- bar
 |   |   |-- hoo
 |   |   |   |-- dar
 |   |   |   |-- far
 |   |   |   |-- gar
 |   |   |   |-- har
 |   |   |-- joo
 |   |   |-- koo
 |   |-- car
 |-- goo"""

class TreeTestCase(unittest.TestCase):
    def setUp(self):
        self.t = tree.Tree()

    def test_add_nodes(self):
        """ Add nodes to a path """
        expected = {
            'foo': {
                'bar': {
                    'goo': None,
                    'car': None
                    }
            }}
        self.t.add_nodes('/foo/bar', ['goo', 'car'])
        self.assertEqual(expected, self.t.nodes)

    def test_simple_tree(self):
        """ Stringify ourselves. """
        self.t.nodes = {
            'foo': {
                'bar': None,
                'car': None
                },
            'goo': None
            }
        self.assertEqual(SIMPLE_TREE, self.t.tree())

    def test_nested_tree(self):
        "Printed representation of a nested tree"
        self.t.nodes = {
            'foo': {
                'bar': {
                    'hoo': {
                        'dar': None,
                        'far': None,
                        'gar': None,
                        'har': None
                        },
                    'joo': None,
                    'koo': None
                    },
                'car': None
                },
            'goo': None
            }
        self.assertEqual(NESTED_TREE, self.t.tree())

    def test_pprint(self):
        "Test our Printing"
        with patch.object(self.t, 'tree') as Psc:
            Psc.return_value = ''
            self.t.pprint()
            Psc.assert_called_once_with()

    def tearDown(self):
        pass



if __name__ == '__main__':
    unittest.main()



"""
unittests for the zoop.client module
"""
import sys
import unittest
if sys.version_info < (2, 7):
    import unittest2 as unittest

from mock import patch, Mock
import zookeeper

import zoop
from zoop import client, exceptions, logutils, queue

logutils.set_loglevel('ERROR')

class BaseZKTestCase(unittest.TestCase):
    def setUp(self):
        self.zk = client.BaseZK('localhost:2181')

    def test_init(self):
        """ store vars """
        self.assertEqual('localhost:2181', self.zk.server)
        self.assertEqual(None, self.zk._zk)

    def test_connect(self):
        """ Connect to the Zookeeper instance """
        self.zk.connwait = 0.1

        with patch.object(client.zookeeper, 'init') as Pinit:

            def connected(server, handler, *args):
                handler(1, -1, 1, None)
                return 1

            Pinit.side_effect = connected
            with patch.object(self.zk, 'watcher'):
                self.assertEqual(None, self.zk._zk)
                self.zk.connect()
                Pinit.assert_called_once()
                self.assertEqual(1, self.zk._zk)
                self.assertEqual(True, self.zk.connected)

    def test_queue_no_connection(self):
        """ Should raise an error """
        with self.assertRaises(exceptions.NotConnectedError):
            self.zk.Queue('/myq')

    def test_queue(self):
        self.zk.connected = True
        with patch.object(queue, 'Queue') as Pq:
            q = self.zk.Queue('/myq')
            Pq.assert_called_once_with(self.zk, '/myq', prefix='q-')


class ClientTestCase(unittest.TestCase):
    def setUp(self):
        self.zk = client.ZooKeeper('localhost:2181')

    def test_create(self):
        """ Create a node """
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.create('/foo/bar', 'YAY')
            self.assertEqual(Pzk.create.return_value, resp)
            Pzk.create.assert_called_once_with(self.zk._zk, '/foo/bar', 'YAY',
                                               [client.OPEN_ACL_UNSAFE], 0)

    def test_create_exists(self):
        """ Raise if it exists """
        def raiser(*a, **k):
            raise(zookeeper.NodeExistsException("!"))

        with patch.object(client.zookeeper, 'create') as Pcreate:
            Pcreate.side_effect = raiser
            with self.assertRaises(exceptions.NodeExistsError):
                self.zk.create('/exists')

    def test_delete(self):
        """ Delete a node """
        with patch.object(client, 'zookeeper') as Pzk:
            self.zk.delete('/foo/bar')
            Pzk.delete.assert_called_once_with(self.zk._zk, '/foo/bar')

    def test_delete_no_node(self):
        """ Delete a node """
        def raiser(*a,**kw):
            raise(zookeeper.NoNodeException("!"))

        with patch.object(client.zookeeper, 'delete') as Pdel:
            Pdel.side_effect = raiser

            with self.assertRaises(exceptions.NoNodeError):
                self.zk.delete('/foo/bar')
                Pdel.assert_called_once_with(self.zk._zk, '/foo/bar')

    def test_get(self):
        """ Should Make a get request to libzookeeper """
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.get('/foo/bar')
            self.assertEqual(Pzk.get.return_value, resp)
            Pzk.get.assert_called_once_with(self.zk._zk, '/foo/bar', None)

    def test_get_watch(self):
        """ Should Make a get request to libzookeeper """
        with patch.object(client, 'zookeeper') as Pzk:
            watch = Mock(name='Mock Watch')
            resp = self.zk.get('/foo/bar', watch)
            self.assertEqual(Pzk.get.return_value, resp)
            Pzk.get.assert_called_once_with(self.zk._zk, '/foo/bar', watch)


    def test_get_children(self):
        """ Should Make a get_children request to libzookeeper """
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.get_children('/foo/bar')
            self.assertEqual(Pzk.get_children.return_value, resp)
            Pzk.get_children.assert_called_once_with(self.zk._zk, '/foo/bar', None)

    def test_exists(self):
        """ Should Make an exists request to libzookeeper """
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.exists('/foo/bar')
            self.assertEqual(Pzk.exists.return_value, resp)
            Pzk.exists.assert_called_once_with(self.zk._zk, '/foo/bar', None)

    def test_exists_watch(self):
        """ Should Make an exists request to libzookeeper """
        cb = Mock(name='Mock Callback')
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.exists('/foo/bar', watch=cb)
            self.assertEqual(Pzk.exists.return_value, resp)
            Pzk.exists.assert_called_once_with(self.zk._zk, '/foo/bar', cb)

    def test_ls(self):
        """ Replicate ls """
        with patch.object(client, 'zookeeper') as Pzk:
            resp = self.zk.ls('/goo/car')
            Pzk.get_children.assert_called_once_with(self.zk._zk, '/goo/car', None)
            self.assertEqual(Pzk.get_children.return_value, resp)

    def test_rmrf(self):
        """ Remove recursive """
        rvals = ['child2', 'child']
        def rets(*a,**kw ):
            if rvals:
                return [rvals.pop()]
            else:
                return []

        with patch.object(client, 'zookeeper') as pzk:
            pzk.get_children.side_effect = rets
            self.zk.rm_rf('/foo/bar/baz')
            pzk.delete.assert_any_call(self.zk._zk, '/foo/bar/baz/child/child2')
            pzk.delete.assert_any_call(self.zk._zk, '/foo/bar/baz/child')
            pzk.delete.assert_any_call(self.zk._zk, '/foo/bar/baz')

    def test_watch(self):
        """ Register our desire to watch for events """
        cb = lambda *a,**k: True
        with patch.object(self.zk.watcher, 'spyon') as Pspy:
            self.zk.watch('/foo/bar', zoop.Event.Child, cb)
            Pspy.assert_called_once_with('/foo/bar', zoop.Event.Child, cb)

    def test_queue_no_connection(self):
        """ Should raise an error """
        with self.assertRaises(exceptions.NotConnectedError):
            self.zk.Queue('/myq')

    def test_queue(self):
        self.zk.connected = True
        with patch.object(queue, 'Queue') as Pq:
            q = self.zk.Queue('/myq')
            Pq.assert_called_once_with(self.zk, '/myq', prefix='q-')

class AsyncClientTestCase(unittest.TestCase):
    def setUp(self):
        self.zk = client.AsyncZooKeeper('localhost:2181')

    def test_create(self):
        """ Create a node """
        pass




if __name__ == '__main__':
    unittest.main()



import re

from distutils.core import setup

VERSION_FILE = "zoop/_version.py"
verstrline = open(VERSION_FILE, "rt").read()
VSRE = r'^__version__ = [\'"]([^\'"]*)[\'"]'
mo = re.search(VSRE,  verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in {0}".format(VERSION_FILE))

setup(
    name = "zoop",
    version = VERSION,
    author = "David Miller",
    author_email = "david@deadpansincerity.com",
    url = "http://github.com/davidmiller/zoop",
    description = "Pythonic API and common paradigms for Zookeeper",
    long_description = open('README.rst').read() + "\n\n" +  open('HISTORY.rst').read(),
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2.6",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries"
        ],
    packages = ['zoop'],
    )



"""
Unittests for the zoop.watch module
"""
import sys
import unittest
if sys.version_info < (2, 7):
    import unittest2 as unittest

from mock import patch, Mock

import zoop
from zoop import enums, exceptions, watch

class WatcherTestCase(unittest.TestCase):
    def setUp(self):
        self.w = watch.Watcher(2)
        self.mock_callback = Mock(name='Mock Callback')

    def test_sethandle(self):
        "Attribute setter for interface"
        mock_handle = Mock(name='Mock Handle')
        self.w.set_zhandle(mock_handle)
        self.assertTrue(self.w._zk is mock_handle)

    def test_dispatch(self):
        """ Run our callbacks and reregister them """
        cb = Mock(name='Mock Callback')
        mock_kids = Mock(name='Mock get_children()')
        values = {enums.Event.Child: mock_kids}
        self.w.callbacks['/foo/bar'][enums.Event.Child].append(cb)
        with patch.dict(watch.Watcher._watch_funcs, values):
            self.w.dispatch(2, enums.Event.Child, None, '/foo/bar')
            cb.assert_called_once_with('/foo/bar', enums.Event.Child)
            args = mock_kids.call_args[0]
            self.assertEqual(self.w._zk, args[0])
            self.assertEqual('/foo/bar', args[1])

    def test_spyon_no_events(self):
        """ Raise when no events passed """
        with self.assertRaises(exceptions.NoEventError):
            self.w.spyon('/joo/kar', self.mock_callback)

    def test_spyon_child(self):
        """ Register our desire to watch for events """
        cb = lambda *a,**k: True
        mock_kids = Mock(name='Mock get_children()')
        values = {enums.Event.Child: mock_kids}
        with patch.dict(watch.Watcher._watch_funcs, values):
            self.w.spyon('/foo/bar', cb, zoop.Event.Child)
            self.assertEqual([cb], self.w.callbacks['/foo/bar'][zoop.Event.Child])
            args = mock_kids.call_args[0]
            self.assertEqual(self.w._zk, args[0])
            self.assertEqual('/foo/bar', args[1])

    def test_spyon_changed(self):
        """ Register our desire to watch for events """
        cb = lambda *a,**k: True
        mock_get = Mock(name='Mock get()')
        values = {
            enums.Event.Changed: mock_get,
            enums.Event.Deleted: mock_get
            }

        with patch.dict(watch.Watcher._watch_funcs, values):
            self.w.spyon('/foo/bar', cb, zoop.Event.Changed, zoop.Event.Deleted)
            self.assertEqual([cb], self.w.callbacks['/foo/bar'][zoop.Event.Changed])
            self.assertEqual([cb], self.w.callbacks['/foo/bar'][zoop.Event.Deleted])
            args = mock_get.call_args[0]
            self.assertEqual(self.w._zk, args[0])
            self.assertEqual('/foo/bar', args[1])



if __name__ == '__main__':
    unittest.main()



"""
unittests for the zoop.client module
"""
import os
import sys
import unittest
if sys.version_info < (2, 7):
    import unittest2 as unittest

from mock import Mock
import zookeeper

from zoop import exceptions, queue

class QueueTestCase(unittest.TestCase):
    def setUp(self):
        self.zk = Mock(name='Mock ZooKeeper')
        self.q = queue.Queue(self.zk, '/foo/q')

    def test_init_exists(self):
        "Don't create, it exists"
        zk = Mock(name='Mock ZooKeeper')
        zk.exists.return_value = True
        queue.Queue(zk, '/bar/q')
        zk.exists.assert_any_call('/bar/q')

    def test_init_doesnt_exist(self):
        "Don't create, it exists"
        zk = Mock(name='Mock ZooKeeper')
        zk.exists.return_value = False
        queue.Queue(zk, '/bar/q')
        zk.exists.assert_any_call('/bar/q')
        zk.create.assert_any_call('/bar/q')

    def test_empty(self):
        """ The queue is empty """
        self.zk.get_children.return_value = []
        self.assertEqual(True, self.q.empty())

    def test_not_empty(self):
        """ The Queue is not empty """
        self.zk.get_children.return_value = ['q-1', 'q-2']
        self.assertEqual(False, self.q.empty())

    def test_flush(self):
        "Flush a Q"
        nodes = ['q-1', 'q-2']
        self.zk.get_children.return_value = nodes
        self.q.flush()
        self.zk.get_children.assert_called_once()
        for n in nodes:
            self.zk.delete.assert_any_call(os.path.join('/foo/q', n))

    def test_get(self):
        """ Take the next element from the Queue"""
        self.zk.get_children.return_value = ['q-1', 'q-2']
        self.zk.get.return_value = 'Q1 Data'

        self.assertEqual("Q1 Data", self.q.get())

        self.zk.get_children.assert_called_once_with('/foo/q')
        self.zk.get.assert_called_once_with('/foo/q/q-1')
        self.zk.delete.assert_called_once_with('/foo/q/q-1')

    def test_get_empty(self):
        "The Queue is empty, raise an Empty error"
        self.zk.get_children.return_value = []
        with self.assertRaises(exceptions.Empty):
            self.q.get()

    def test_put(self):
        """ Put an item into the Queue """
        self.q.put('Foo')
        self.zk.create.assert_called_once_with('/foo/q/q-', value='Foo', flags=zookeeper.SEQUENCE)

    def test_qsize(self):
        "Length of the Q"
        self.zk.get_children.return_value = ['q-1']
        self.assertEqual(1, self.q.qsize())

    def test_sorted(self):
        """ Get the sorted list of items """
        self.zk.get_children.return_value = ['q-2', 'q-4', 'q-3']
        self.assertEqual(['q-2', 'q-3', 'q-4'], self.q.sorted())

    def test_sorted_empty(self):
        self.zk.get_children.return_value = []
        with self.assertRaises(exceptions.Empty):
            self.q.sorted()

    def test_watch(self):
        cb = Mock(name='Mock Callback')
        self.q.watch(cb)
        self.assertEqual(True, self.zk.watch.called)

    def test_watchitem(self):
        cb = Mock(name='Mock Callback')
        self.q.watchitem(cb)
        self.assertEqual(True, self.zk.watch.called)

    def tearDown(self):
        pass



if __name__ == '__main__':
    unittest.main()
